package connector

import (
	"context"
	"errors"
	"fmt"
	"io"
	"net"
	"strings"
	"time"

	"golang.org/x/crypto/ssh"

	"gitlab.com/gitlab-org/fleeting/fleeting/provider"
)

type sshClient struct {
	client *ssh.Client
}

const (
	// handshakeErrorRetries is the number of retries performed specifically
	// for "handshake"/initial temporary connection related errors.
	//
	// These occur often when a new instance is being created because the SSH
	// socket might be available, but setup via init scripts are still
	// configuring and/or restarting the daemon.
	//
	// 30 attempts and at a 2 second interval is pretty long for this
	// specific type of problem, however, it's also capped by the context
	// deadline and the timeout provided for it. So it being too long is
	// better than it being too short.
	handshakeErrorRetries       = 30
	handshakeErrorRetryInterval = 2 * time.Second
)

func DialSSH(ctx context.Context, info provider.ConnectInfo, options DialOptions) (*sshClient, error) {
	addr := info.InternalAddr
	if options.UseExternalAddr && (info.ExternalAddr != "" || addr == "") {
		addr = info.ExternalAddr
	}
	addr = hostport(addr, "22")

	config := &ssh.ClientConfig{
		User:            info.Username,
		HostKeyCallback: ssh.InsecureIgnoreHostKey(),
		Timeout:         info.Timeout,
	}

	if info.Key == nil {
		config.Auth = append(config.Auth, ssh.Password(info.Password))
	} else {
		signer, err := ssh.ParsePrivateKey(info.Key)
		if err != nil {
			return nil, err
		}
		config.Auth = append(config.Auth, ssh.PublicKeys(signer))
	}

	var client *ssh.Client
	var err error
	i := 0
	for ; i < handshakeErrorRetries; i++ {
		client, err = getSSHClient(ctx, dialer{
			Timeout:   info.Timeout,
			KeepAlive: info.Keepalive,
			DialFn:    options.DialFn,
		}, addr, config)
		if err != nil && isTemporarySSHError(err) {
			time.Sleep(handshakeErrorRetryInterval)
			continue
		}
		break
	}
	if err != nil {
		return nil, fmt.Errorf("after retrying %d times: %w", i, err)
	}

	return &sshClient{client: client}, nil
}

func isTemporarySSHError(err error) bool {
	str := err.Error()

	return strings.Contains(str, "handshake failed") || strings.Contains(str, "i/o timeout")
}

func (c *sshClient) Close() error {
	return c.client.Close()
}

func (c *sshClient) Run(ctx context.Context, opts RunOptions) error {
	sess, err := c.client.NewSession()
	if err != nil {
		return err
	}
	defer sess.Close()

	sess.Stdin = opts.Stdin
	sess.Stdout = opts.Stdout
	sess.Stderr = opts.Stderr

	errCh := make(chan error, 1)

	go func() {
		errCh <- sess.Run(opts.Command)
	}()

	select {
	case <-ctx.Done():
		return sess.Close()
	case err := <-errCh:
		return c.newExitError(err)
	}
}

func (c *sshClient) newExitError(err error) error {
	if err == nil {
		return nil
	}

	var sshExitErr *ssh.ExitError
	if errors.As(err, &sshExitErr) {
		return &ExitError{
			err:      err,
			exitCode: sshExitErr.ExitStatus(),
		}
	}

	return &ExitError{err: err, exitCode: 0}
}

func (c *sshClient) DialRun(ctx context.Context, command string) (net.Conn, error) {
	connReader, w := io.Pipe()
	r, connWriter := io.Pipe()

	go func() {
		ctx, cancel := context.WithCancel(ctx)
		defer cancel()

		defer r.Close()
		defer w.Close()

		stderr := newStderrOmitWriter()
		err := c.Run(ctx, RunOptions{
			Command: command,
			Stdin:   r,
			Stdout:  w,
			Stderr:  stderr,
		})
		if err != nil {
			err = fmt.Errorf("winrm tunnel: %w (%s)", err, stderr.Error())
			r.CloseWithError(err)
			w.CloseWithError(err)
		}
	}()

	return &rwConn{connWriter, connReader}, nil
}

func (c *sshClient) Dial(network string, address string) (net.Conn, error) {
	return c.client.Dial(network, address)
}

func getSSHClient(ctx context.Context, d dialer, address string, config *ssh.ClientConfig) (*ssh.Client, error) {
	ctx, cancel := context.WithCancel(ctx)
	defer cancel()

	conn, err := d.DialContext(ctx, "tcp", address)
	if err != nil {
		return nil, err
	}

	c, chans, reqs, err := ssh.NewClientConn(conn, address, config)
	if err != nil {
		return nil, err
	}

	return ssh.NewClient(c, chans, reqs), nil
}
